#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <string.h>

// Socket related include files
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/time.h>
#include "Functions.h"
#include "Message.h"

extern char* 	sys_errlist[] ; 

const int	hostNameRetries = 7 ;	
const int	STR_IP_ADDRESS	= 1 ;
const int	STR_HOST_NAME   = 2;


// -----------------------------------------------------------------------------------
// OurGetHostByName -- This is due to the slowness of the name binding software. If it
//                     fails for the first name, then it might be due to non caching
//		       of names.  	 
//                     The number of times we retry is adhoc
// -----------------------------------------------------------------------------------
struct hostent*
OurGetHostByName( const char* hostName )
{
	register struct hostent*	tmpHostEnt ;

	assert( hostName != NULL ) ;

	for ( int i = 0 ; i < hostNameRetries ; i++ )
	{
		tmpHostEnt = gethostbyname( hostName ) ;

		if ( tmpHostEnt )
		{
			return tmpHostEnt ;
		}
	}

	return NULL ;
}

// -----------------------------------------------------------------------------------
// OurGetHostByAddr -- This is due to the slowness of the name binding software. If it
//                     fails for the first name, then it might be due to non caching
//		       of names.  	 
//                     The number of times we retry is adhoc
// -----------------------------------------------------------------------------------
struct hostent*
OurGetHostByAddr( const char* hostAddr, int Length, int addrType )
{
	register struct hostent*	tmpHostEnt ;

	assert( hostAddr != NULL ) ;

	for ( int i = 0 ; i < hostNameRetries ; i++ )
	{
		tmpHostEnt = gethostbyaddr( hostAddr, Length, addrType ) ;

		if ( tmpHostEnt )
		{
			return tmpHostEnt ;
		}
	}

	return NULL ;
}

int
CheckIPAddress( const char* hostNameORAddr )
{
	assert( hostNameORAddr != NULL ) ;

	signed long  flag = inet_addr ( hostNameORAddr )  ;
	
	if ( flag < 0 ) 
	{
		struct hostent* hostEnt = OurGetHostByName(hostNameORAddr) ;

		if ( hostEnt  == NULL )
			return -1 ;

		return STR_HOST_NAME ;	
	}

	return STR_IP_ADDRESS ;
}

struct hostent*
OurGetHostEntry( const char*  hostNameORAddr )
{
	assert( hostNameORAddr != NULL ) ;

	int addrHost = CheckIPAddress(hostNameORAddr) ;

	if ( addrHost == STR_IP_ADDRESS )
	{
		struct sockaddr_in	hostAddr ;

		bzero( (char *)&hostAddr, sizeof(struct sockaddr_in) ) ; 

		hostAddr.sin_addr.s_addr = inet_addr( hostNameORAddr ) ;

		return OurGetHostByAddr( (char *)&hostAddr, sizeof(hostAddr), AF_INET ) ; 
	}

	if ( addrHost == STR_HOST_NAME )
	{
		return OurGetHostByName( hostNameORAddr ) ;
	}

	return NULL ;
}

char*
OurGetHostName( const char* hostNameORAddr )
{
	static char hostName[MAXHOSTNAMELEN] ;

	assert( hostNameORAddr != NULL ) ;

	int addrHost = CheckIPAddress(hostNameORAddr) ;

	struct hostent*	hostEntry ; 

	if ( addrHost == STR_IP_ADDRESS )
	{
		struct sockaddr_in 	hostAddr ;

		bzero( (char *)&hostAddr, sizeof(struct sockaddr_in) ) ; 

		hostAddr.sin_addr.s_addr = inet_addr( hostNameORAddr ) ;

		hostEntry = OurGetHostByAddr( (char *)&hostAddr, 
					sizeof(hostAddr), AF_INET ) ; 

		::strncpy(hostName, hostEntry->h_name, MAXHOSTNAMELEN - 2 ) ;

		hostName[::strlen(hostName) > MAXHOSTNAMELEN ? 
			MAXHOSTNAMELEN - 1 : strlen(hostName) ] = '\0' ;

		return hostName ;

	}

	if ( addrHost == STR_HOST_NAME )
	{
		::strncpy( hostName, hostNameORAddr, MAXHOSTNAMELEN - 2 ) ;

		hostName[::strlen(hostName) > MAXHOSTNAMELEN ? 
			MAXHOSTNAMELEN - 1 : strlen(hostName) ] = '\0' ;

		return hostName ; 
	}

	return NULL ;
}

char*
OurGetHostAddress( const char* hostNameORAddr )
{
	static char hostAddr[MAXHOSTNAMELEN] ;

	register struct in_addr* inAddrEnt = NULL ;

	assert( hostNameORAddr != NULL ) ;

	int addrHost = CheckIPAddress(hostNameORAddr) ;

	struct hostent*	hostEntry ; 

	if ( addrHost == STR_IP_ADDRESS )
	{
		::strcpy(hostAddr,hostNameORAddr ) ;

		return hostAddr ;
	}

	if ( addrHost == STR_HOST_NAME )
	{
		hostEntry = OurGetHostByName( hostNameORAddr ) ;

		inAddrEnt = (struct in_addr *) *(hostEntry->h_addr_list) ;

		// Return the first IP address
		return inet_ntoa(*inAddrEnt) ;
	}

	return NULL ;
}

char*
GetHostName()
{
	static char hostName[MAXHOSTNAMELEN] ;

	for ( int i = 0 ; i < hostNameRetries ; i++ )
	{
		if ( gethostname( hostName,sizeof(hostName) ) == 0 ) ;
			return hostName ;
	}

	return NULL ;
}


int
CreateServerSocket( int portNumber )  
{
	register int sockFd ;

	struct sockaddr_in servAddr ;  

	sockFd = socket(AF_INET,SOCK_STREAM,0) ;

	if ( sockFd < 0 )
	{
		cerr << "Error creating server socket:" << sys_errlist[errno] << endl ;
		return -1 ;
	}


	bzero( (char *)&servAddr, sizeof(struct sockaddr_in) ) ;  

	servAddr.sin_family = AF_INET ;
	servAddr.sin_addr.s_addr = htonl(INADDR_ANY) ;
	servAddr.sin_port = htonl(portNumber) ;

	if ( bind(sockFd,(struct sockaddr *)&servAddr, sizeof(servAddr) ) < 0 )
	{
		cerr << "Error binding server socket:" << sys_errlist[errno] << endl ;
		close(sockFd) ;
		return -1 ; 
	}

	if ( listen(sockFd,5) < 0 )
	{
		cerr << "Error in listening to socket:" << sys_errlist[errno] << endl ; 
		return -1 ;
	}

	return sockFd ;
}

// -----------------------------------------------------------------------------------
// CreateClientSocket - Creates a client socket for communication with GSP. 
//                      Has to be converted to be generic so that the parameter
//                      hostName is either a real host name or dotted ip address
// -----------------------------------------------------------------------------------
int
CreateClientSocket(const char* hostNameORAddr, int servPortNumber )
{
	register int 			sockFd ;
	register struct hostent* 	servHostEnt ;
	register struct in_addr* 	inAddrEnt ;
	struct sockaddr_in 		servAddr ;  

	int	 			error ;

 	assert( hostNameORAddr != NULL ) ;

	sockFd = socket(AF_INET,SOCK_STREAM,0) ;

	if ( sockFd < 0 )
	{
		cerr << "Error creating client socket:" << sys_errlist[errno] << endl ;
		return -1 ;
	}

	bzero( (char *)&servAddr, sizeof(servAddr) ) ;  

	// Assuming the host name can be either a string or string of ip address
	// in dotted notation

	servHostEnt = OurGetHostEntry( hostNameORAddr ) ;

	if ( !servHostEnt )
	{
		cerr << "Error getting host entry for " << hostNameORAddr << endl ;
		close(sockFd) ;
		return -1 ;
	}	

	servAddr.sin_family = AF_INET ;
	servAddr.sin_port = htonl(servPortNumber) ;

	// Loop thru' the host entry address list since it is possible
	// some GSP machines have multiple IP addresses

	inAddrEnt = (struct in_addr *) *(servHostEnt->h_addr_list) ;

	while ( inAddrEnt )
	{
		// Try connecting the current iterating address 
		servAddr.sin_addr.s_addr = inet_addr( inet_ntoa(*inAddrEnt) ) ;

		error = connect(sockFd, (struct sockaddr *)&servAddr,sizeof(servAddr)) ; 

		// Successfully connected 
		if ( error >= 0 ) 
		{
			// Set the linger option so that the data
			// gets flushed on socket close
			struct linger lingerOnClose = {1,1} ;

			setsockopt( sockFd, SOL_SOCKET, SO_LINGER, 
				(char *)&lingerOnClose, sizeof(struct linger) ) ;

			return sockFd ;
		}	

		cerr << "Error connecting to " << servHostEnt->h_name << ':' << sys_errlist[errno] << endl ;
		// try other addresses  
		servHostEnt->h_addr_list++ ;

		inAddrEnt = (struct in_addr *) *(servHostEnt->h_addr_list) ;
	}

	cerr << "Error connecting to " << servHostEnt->h_name << ':' << sys_errlist[errno] << endl ;

	close(sockFd) ;

	return -1 ;	
}
