#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <ctype.h>

// Socket related include files
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "Constant.h"
#include "Functions.h"

// C = A - B 
// Warning time zone information is neglected
// Assumes that A is the latest time
void
DiffTime( struct timeb& timeA, struct timeb& timeB, struct timeb& timeC )
{
	timeC.millitm = timeA.millitm - timeB.millitm + 
			(timeB.millitm > timeA.millitm ? 1000 : 0) ; 
			
	timeC.time = timeA.time - timeB.time - ( timeB.millitm > timeA.millitm ? 1 : 0 ) ;
}
			

// -----------------------------------------------------------------------------------------
// CheckValidInteger -- This function takes a string and checks whether it is an integer. 
//	                Returns -1 when the string is not a valid integer.
//			Returs 0 when the string is a valid integer. 
// -----------------------------------------------------------------------------------------

int
CheckValidInteger( const char* numString )
{
	assert( numString != NULL ) ;

	if ( !numString )
		return  -1 ;

	while (*numString && isdigit(*numString++) ) ;

	return isdigit(*--numString) ? 0 : -1 ; 
}

// ------------------------------------------------------------------------------------------
// CheckFileExistence -- This function takes a file name and checks whether it exists in the
// 			 file system or not
//			 Returns 0 when the file exists
//			 Returns -1 otherwise 
// ------------------------------------------------------------------------------------------

int
CheckFileExistence( const char* fileName )
{
	assert ( fileName != NULL ) ;

	if ( !fileName )
		return -1 ;

	struct stat buffer ;

	if ( stat( fileName,&buffer) < 0 )   
	{
		return -1 ;
	}

	return 0 ; 
}

int
ValidatePortNumber ( const char* portNumber, int selection )
{
	if ( !portNumber )
	{
		switch ( selection )
		{
			case GSP_PORT_NUMBER :
		
				return Constant::GSPPortNumber ;
				
			case LSP_PORT_NUMBER : 

				return Constant::LSPPortNumber ;

			case USER_PORT_NUMBER :

				return Constant::UserPortNumber ; 

			case QUERY_PORT_NUMBER :

				return Constant::QueryPortNumber ; 
		}
	}
	else if ( !CheckValidInteger( portNumber ) ) 
	{
		return atoi(portNumber) ;
	}
		
	return -1 ;
}

// -----------------------------------------------------------------------------------------
// GetConfigurationInfo -- This function reads the given configuration file, scans for
//			   the given hostaddress( input can be either hostName or IP address
//			   and returns the number of processors it has.
// -----------------------------------------------------------------------------------------

int
GetConfigurationInfo( const char* configFileName, const char* hostNameORAddr )
{
	assert( configFileName != NULL ) ;

	assert( hostNameORAddr != NULL ) ;

 	FILE* fp = fopen(configFileName,"r") ;

	if ( !fp )
	{
		cerr << "Error opening configuration file " << configFileName << endl ;

		return -1 ;
	}

	// The function is written in a crude way we should have a parser kind of
	// stuff which can detect the syntax and accordingly tell the errors

	char hostAddress[BUFSIZ] ; 

	char noOfProcessors[BUFSIZ] ;

	register struct	hostent*	hostEntry ;

	register struct in_addr*        inAddrEnt ;

	hostEntry = OurGetHostEntry( hostNameORAddr ) ;

	if ( !hostEntry )
	{
		cerr << "Error getting host entry for " << hostNameORAddr << endl ; 
		return -1 ;
	}

	while ( !feof(fp) )
	{
		fscanf(fp,"%s %s\n",hostAddress,noOfProcessors ) ;

		// Search for its alternate addresses also
		// This is to support multiported machines

		inAddrEnt = (struct in_addr *) *(hostEntry->h_addr_list) ;

		register char** tmpAddrList = hostEntry->h_addr_list ;

		while ( inAddrEnt ) 
		{
			if ( !::strcmp( inet_ntoa(*inAddrEnt), hostAddress ) )
			{	
				fclose(fp) ;

				return !CheckValidInteger(noOfProcessors) ?
					atoi(noOfProcessors) : -1 ;
			}

			tmpAddrList++ ;

			inAddrEnt = (struct in_addr *) *tmpAddrList ;
		}
	}

	cerr << "Entry for host " << hostNameORAddr 
	     << " is not found in conf file " << configFileName << endl ;

	return -1 ; 
}

// -----------------------------------------------------------------------------------------
// GetIPAddress -- This function reads the given configuration file, scans for
//			   the given hostname
//			   and returns the ip address the GSP uses.
// -----------------------------------------------------------------------------------------

int
GetIPAddress( const char* configFileName, const char* hostName, char* hostAddr )
{
	assert( configFileName != NULL ) ;

	assert( hostName != NULL ) ;

	assert( hostAddr != NULL ) ;

 	FILE* fp = fopen(configFileName,"r") ;

	if ( !fp )
	{
		cerr << "Error opening configuration file " << configFileName << endl ;

		return -1 ;
	}

	// The function is written in a crude way we should have a parser kind of
	// stuff which can detect the syntax and accordingly tell the errors

	char hostAddress[BUFSIZ] ; 

	char noOfProcessors[BUFSIZ] ;

	register struct	hostent*	hostEntry ;

	register struct in_addr*        inAddrEnt ;

	hostEntry = OurGetHostEntry( hostName ) ;

	if ( !hostEntry )
	{
		cerr << "Error getting host entry for " << hostName << endl ; 
		return -1 ;
	}

	while ( !feof(fp) )
	{
		fscanf(fp,"%s %s\n",hostAddress,noOfProcessors ) ;

		// Search for its alternate addresses also
		// This is to support multiported machines

		inAddrEnt = (struct in_addr *) *(hostEntry->h_addr_list) ;

		register char** tmpAddrList = hostEntry->h_addr_list ;

		while ( inAddrEnt ) 
		{
			if ( !::strcmp( inet_ntoa(*inAddrEnt), hostAddress ) )
			{	
				fclose(fp) ;

				::strcpy( hostAddr, hostAddress ) ; 

				return 0 ;
			}

			tmpAddrList++ ;

			inAddrEnt = (struct in_addr *) *tmpAddrList ;
		}
	}

	cerr << "Entry for host " << hostName 
	     << " is not found in conf file " << configFileName << endl ;

	return -1 ; 
}
