#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

// Socket related include files
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/time.h>
#include <sys/timeb.h>
#include "Functions.h"
#include "Message.h"
#include "List.h"
#include "GSPJob.h"
#include "GSPMachine.h"
#include "queue.h"

job_queue	policies ;

const	int		MAX_SOCKETS_SUPPORTED = 10 ; 
int			currPolicy = PRIORITY;		// This is FIFO

// Locals
char			MyIPAddress[MAXHOSTNAMELEN] ;

String			submitDir ;						
String			HostConfigFile ;	
int			GSPServerSock[MAX_SOCKETS_SUPPORTED] ; 
int			GSPportNumber ; 
int 			UserportNumber ;
int 			QueryportNumber ;
int			GSPUserSock ;
int			GSPQuerySock ;

oo_List<Job_t>		jobList ;
HashTable_t		machineList ;
oo_List<RunningJob_t>	runningList ;

// Logging job completion details
ofstream		fJobLog ;

int  
BuildAndSendScheduleMessage ( Job_t* ) ;

int
ProcessSubmit ( int ) ;

int
ProcessTerminate( int, const char*, int, struct timeb& ) ;  

int
SingleGSPJobQuery( int, int, char*) ;

int
AllGSPJobQuery( int, char*) ;
	
int
KillAJob( int ) ;

int
SuspendJob( int ) ;

int
ResumeJob( int ) ;

int 
Initialize()
{
	char*	myName = NULL ;

	// Now get my host name
	if ( !(myName = GetHostName()) )
	{
		cerr << "Problem getting my host name. " << endl ;
		exit(1) ;
	}

	// Now get my ip address. this will succeed only if there is
	// an entry for this machine in configuration file
	if ( GetIPAddress( HostConfigFile.chars(), myName, MyIPAddress ) < 0 )
		exit(1) ;

	// Build the hash table of the machines in the configuration file 
	int socks = BuildMachineList( machineList, HostConfigFile.chars() ) ;

	if ( socks < 0 )
	{
		cerr << "Unable to read host configuration file." << endl ; 

		return -1 ;
	}

	for ( int i = 0 ; i < MAX_SOCKETS_SUPPORTED ; i++ )
	{
		// Create server socket for lsps to send messages
		GSPServerSock[i] = CreateServerSocket( GSPportNumber+i ) ;

		if ( GSPServerSock[i] < 0 )
		{
			cerr << "Unable to create server socket to listen in port "
		     		<< GSPportNumber+i << endl ; 

			return -1 ; 	
		}

	}

	// Create server socket for GSP Queries
	GSPQuerySock = CreateServerSocket( QueryportNumber ) ;

	if ( GSPQuerySock < 0 )
	{
		cerr << "Unable to create server socket to listen in port "
		     << QueryportNumber << endl ; 

		return -1 ; 	
	}

	// Create server socket for job submission
	GSPUserSock = CreateServerSocket( UserportNumber ) ;

	if ( GSPUserSock < 0 )
	{
		cerr << "Unable to create server socket to listen in port "
		     << UserportNumber << endl ; 

		return -1 ; 	
	}

	return 0 ; 
}

int
AllocateNodesForAJob( Job_t* ) ;

void
signal_INT(int x)
{
	close(GSPUserSock) ;

	for ( int i = 0 ; i < MAX_SOCKETS_SUPPORTED ; i++ )
		close(GSPServerSock[i]) ;

	exit(1) ;
}

void
signal_ALRM(int x)
{
}

void
Schedule(Job_t* newJob)
{
	if ( newJob == NULL )
		return ;

	newJob->Parse() ;

	cout << jobList.Length() << endl ;
	
	Print( jobList ) ;

	AllocateNodesForAJob( newJob ) ; 

	Print( runningList ) ; 

	BuildAndSendScheduleMessage( newJob ) ; 
}

void
main(int argc, char* argv[] )
{
	extern	int	optind ;
	extern	char*	optarg ;

	char*	gspPortNumber = NULL ;
	char*	userPortNumber = NULL ;
	char*	queryPortNumber = NULL ;
	char*	conFile = NULL ; 
	char*	myName = NULL ;
	char*	logFile = NULL ; 
	char*	jobDir = NULL ; 
	String	logF ; 

	int	c ;

	while ( (c = getopt(argc,argv,"f:p:u:q:g:d:") ) != EOF )
	{
		 switch(c)
		 {
			case 'f' :
				 conFile = optarg ;
				 break ;

			case 'p' :
				 gspPortNumber = optarg ;
				 break ;
			
			case 'u' :
				 userPortNumber = optarg ;
				 break ;
			
			case 'q' :
				 queryPortNumber = optarg ;
				 break ;

			case 'g' :
				 logFile = optarg ;
				 break ;

			case 'd' :
				 jobDir = optarg ;
				 break ;

			case '?' :
			         cerr << "Usage : " << argv[0] << 
				 " [ -p <GSP Port Number> -u <User Port Number> "
				 << "-q <Query Port Number> ] " 
				 << "-f <Configuration File> ] " 
				 << "-g <Job Log File> ] " 
				 << "-d <Job Submission Directory> ] " << endl ;
				 exit(1) ;
		 }
	}

	// Validate the GSP port number 
	if ( (GSPportNumber = ValidatePortNumber(gspPortNumber,GSP_PORT_NUMBER)) < 0 )
	{
		cerr << "GSP Port Number should be a valid integer" << endl ;		
		exit(1) ;
	}

	// Validate the User port number 
	if ( (UserportNumber = ValidatePortNumber(userPortNumber,USER_PORT_NUMBER)) < 0 )
	{
		cerr << "User Port Number should be a valid integer" << endl ;		
		exit(1) ;
	}

	// Validate the query port number 
	if ( (QueryportNumber = ValidatePortNumber(queryPortNumber,QUERY_PORT_NUMBER)) < 0 )
	{
		cerr << "Query Port Number should be a valid integer" << endl ;		
		exit(1) ;
	}

	// Validate the config file 	
	if ( !conFile )
	{
		HostConfigFile = ConfigFile ;
	}
	else if ( !CheckFileExistence( conFile ) )
	{
		HostConfigFile = conFile ;
	}
	else
	{
		cerr << conFile << " does not exist. Quitting..." << endl ;		

		exit(1) ;
	}

	if ( !logFile )
	{
		logF = JobLogFile ; 
	}
	else
	{
		logF = logFile ; 
	}

	if ( !jobDir )
	{
		submitDir = JobSubmissionDir ; 
	}
	else
	{
		submitDir = jobDir ; 
	}

	fJobLog.open(logF.chars()) ; 
		
	if ( !fJobLog )
	{
		 cerr << "Unable to open job log file. Quitting..." << endl ;

		 exit(1) ;
	}

	if ( Initialize() < 0 )
	{
		cerr << "GSP Initialization Failed. Quitting..." << endl ;      

		exit(1) ;  
	}

	// Install the signal handler for interrupt handling
	signal( SIGINT, signal_INT ) ;

	signal( SIGALRM, signal_ALRM ) ;

	struct sockaddr_in clientAddr ;

	int clientAddrLen = 0 ;

	fd_set readset ;

	int count = 1 ;	

	int tcount = 0 ; 

	int jb = 1 ; 

	while (1)
	{
		Job_t* newJob = NULL;

		FD_ZERO(&readset) ;

		for ( int i = 0 ; i < MAX_SOCKETS_SUPPORTED ; i++ )
		{	
			FD_SET(GSPServerSock[i],&readset) ;
		}

		FD_SET(GSPQuerySock,&readset) ;

		FD_SET(GSPUserSock,&readset) ;

		// GSPUserSock was created last so it will have a higher descriptor

		cout << "Waiting for Messages" << endl ;

		int s = select( GSPUserSock+1 , &readset, NULL, NULL, NULL ) ;

		cout << "After Select s = " << s <<  endl ;

		// Receive on server sockets  
		for ( i = 0 ; i < MAX_SOCKETS_SUPPORTED ; i++ )
		{
			if ( s > 0 && FD_ISSET( GSPServerSock[i], &readset ) )
			{
				int clientD  = accept( GSPServerSock[i], (struct sockaddr *)&clientAddr, &clientAddrLen) ;
	
				if (clientD < 0 )
				{
					cerr << "Error in accept:" << sys_errlist[errno] << endl ;
	
					continue ; 
				}
	
				MessHeader_t msg(MyIPAddress) ;
	
				if ( msg.Receive(clientD) < 0 ) 
				{
					cerr << "Error in receiving message header " 
					     << sys_errlist[errno] << endl ;
	
					close(clientD) ;
	
					continue ; 
				}       
	
				cout << msg << endl ;
	
				if ( msg.msgType == MSG_STARTUP )
				{
					// Activate the appropriate machine
					if ( ActivateMachine( machineList, msg.ipAddress, msg.numNodes ) < 0 )
					{
						cerr << "Spurious startup message received from " << msg.ipAddress
							  << ". Ignoring... " << endl ;
	
						close(clientD) ;
	
						continue ;
					}

					while ( policies.dequeue( currPolicy, machineList.processorsFree, newJob ) )
					{
						Schedule( newJob ) ;
					}
				}
				else if ( msg.msgType == MSG_TERMINATED ) 
				{
					tcount++ ;
	
					cout << "Terminated Count " << tcount << endl ;  
	
					ProcessTerminate( msg.jobId, msg.ipAddress,
					 		  msg.entryNum, msg.timeTakenToComplete ) ; 
				
					// Schedule a new job 	
					while ( policies.dequeue( currPolicy, machineList.processorsFree, newJob ) )
					{
						Schedule( newJob ) ;
					}
				}
				else if ( msg.msgType == MSG_WITHDRAW )
				{
					// Activate the appropriate machine
					if ( DeActivateMachine( machineList, msg.ipAddress, msg.numNodes ) < 0 )
					{
						cerr << "Cannot Withdraw machine " << msg.ipAddress
							  << ". Ignoring... " << endl ;
	
						close(clientD) ;
	
						continue ;
					}
				}

				close(clientD) ; 
			} // end if 
		} // end for 

		if ( s > 0 && FD_ISSET( GSPUserSock, &readset ) )
		{
			int clientS = accept( GSPUserSock, (struct sockaddr *)&clientAddr, &clientAddrLen) ;
			
			if ( clientS < 0 )
			{
				cerr << "Error in accept:" << sys_errlist[errno] << endl ;
				break ;
			}

			MessHeader_t msg(MyIPAddress) ;

			if ( msg.Receive(clientS) < 0 ) 
			{
				cerr << "Error in receiving message header " 
				     << sys_errlist[errno] << endl ;

				close(clientS) ;

				continue ; 
			}       

			cout << msg << endl ;

			if ( msg.msgType == MSG_SUBMIT )
			{
				ProcessSubmit( clientS ) ; 	

				close(clientS) ;

				Job_t*	newJob = NULL ; 

				// Schedule new jobs
				while ( policies.dequeue( currPolicy, machineList.processorsFree, newJob ) )
				{
					Schedule( newJob ) ;
				}
			}
			else
				close(clientS) ;

		}

		if ( s > 0 && FD_ISSET( GSPQuerySock, &readset ) )
		{
			int clientQ = accept( GSPQuerySock, (struct sockaddr *)&clientAddr, &clientAddrLen) ;
			
			if ( clientQ < 0 )
			{
				cerr << "Error in accept:" << sys_errlist[errno] << endl ;
				break ;
			}

			MessHeader_t msg(MyIPAddress) ;

			if ( msg.Receive(clientQ) < 0 ) 
			{
				cerr << "Error in receiving message header " 
				     << sys_errlist[errno] << endl ;

				close(clientQ) ;

				continue ; 
			}       

			cout << msg << endl ;
	
			if ( msg.msgType == MSG_QUERY )
			{
				if ( msg.jobId >= 0 )
				{
					SingleGSPJobQuery( clientQ, msg.jobId, MyIPAddress ) ;
				}
				else if ( msg.jobId < 0 )
				{
					AllGSPJobQuery( clientQ, MyIPAddress ) ;
				}
			}
			else if ( msg.msgType == MSG_KILL )
			{
				if ( KillAJob( msg.jobId ) < 0 )
				{
					cerr << "Unable to kill job " << msg.jobId << endl ;
				}
				else
				{
					while ( policies.dequeue( currPolicy, machineList.processorsFree, newJob ) )
					{
						Schedule( newJob ) ;
					}
				}
			}
			else if ( msg.msgType == MSG_SUSPEND )
			{
				if ( SuspendJob( msg.jobId ) < 0 )
				{
					cerr << "Unable to suspend job " << msg.jobId << endl ;
				}
			}
			else if ( msg.msgType == MSG_RESUME )
			{
				if ( ResumeJob( msg.jobId ) < 0 )
				{
					cerr << "Unable to resume job " << msg.jobId << endl ;
				}
			}


			close(clientQ) ;
		}
		else if ( s < 0 )
		{
			cerr << "Error in select:" << sys_errlist[errno] << endl ;

			break ;
		}

		machineList.Print() ; 	
	}
}

	/*
	int i = 0 ;

	// for ( i = 0 ; i < 100000 ; i++ )
	//{
	Job_t* newJob = new Job_t ;

	newJob->jobId = 5 ;

	newJob->numNodesRequired = 5 ; 

	newJob->jobConfigFile = argv[1] ; 

	ftime( &(newJob->jobArrivalTime) ) ; 

	jobList.Insert( newJob) ;

	newJob->Parse() ;
	//}

	cout << jobList.Length() << endl ;
	
	Print( jobList ) ;

	AllocateNodesForAJob( newJob ) ; 

	Print( runningList ) ; 

	BuildAndSendScheduleMessage( newJob ) ; 
	*/
		/*
		if ( machineList.processorsFree >=4 ) 
		// if ( machineList.processorsFree >=4 && count % 2 ) 
		{

		// Schedule a job 
		Job_t* newJob = new Job_t ;

		newJob->jobId = jb++ ;

		newJob->numNodesRequired = 4 ; 

		newJob->jobConfigFile = "jobfile" ; 

		ftime( &(newJob->jobArrivalTime) ) ; 

		Insert( jobList, newJob) ;

		newJob->Parse() ;

		cout << jobList.Length() << endl ;
	
		Print( jobList ) ;

		AllocateNodesForAJob( newJob ) ; 

		Print( runningList ) ; 

		// alarm(30) ;

		BuildAndSendScheduleMessage( newJob ) ; 

		// alarm(0) ;

		count++ ; 
		}

		machineList.Print() ; 	
		*/
