#include <stdio.h>
#include <fstream.h>
#include <assert.h>

#include "List.h"
#include "GSPJob.h"
#include "GSPMachine.h"

extern	oo_List<Job_t>		jobList ;
extern	oo_List<RunningJob_t>	runningList ;
extern	ofstream		fJobLog	 ;
extern	HashTable_t		machineList ; 	


inline ostream& 
operator << (ostream& os,struct timeb& aTime)
{
	os << aTime.time << "." << aTime.millitm ;

	return os ;
}

int
ProcessCompleted( Job_t* aJob, RunningJob_t* aRunningJob)
{
	assert ( aJob != NULL ) ;

	assert ( aRunningJob != NULL ) ; 

	struct timeb	scheduleOverhead ;
	struct timeb	queueWait ;
	struct timeb	actualRunningTime ;
	double		expectedRunningTime ; 

	// Calculate the scheduling overhead
	DiffTime( aJob->jobFinishScheduleTime, aJob->jobStartScheduleTime, scheduleOverhead ) ;

	// Calculate the waiting time in the queue 
	DiffTime( aJob->jobStartScheduleTime,aJob->jobArrivalTime,queueWait ) ;

	// Calculate the actual running time 
	DiffTime( aJob->jobLatestCompletionTime,aJob->jobFinishScheduleTime, actualRunningTime ) ;

	// Calculate the expected running time
	expectedRunningTime = (double)aJob->expCompletionTime / aJob->numNodesRequired ; 


	fJobLog << aJob->jobId << " " 
		<< scheduleOverhead << " " 
		<< queueWait << " " 
		<< actualRunningTime << " " 
		<< expectedRunningTime << " " 
		<< aJob->numNodesRequired << " " 
		<< aRunningJob->numCompleted << endl ;  

	fJobLog << '*' << " " 
		<< aJob->jobArrivalTime << " " 
		<< aJob->jobStartScheduleTime << " " 
		<< aJob->jobFinishScheduleTime << " " 
		<< aJob->jobEarliestCompletionTime << " " 
		<< aJob->jobLatestCompletionTime << " " 
		<< aJob->numNodesRequired << " " 
		<< aRunningJob->numCompleted << endl ;  

	// Start the machine loop
	// At least a machine will be there in the list since the user
	// cannot submit a job of requirement 0
	MachInfo_t*	tmpMach = NULL ;

	aRunningJob->runningMachList.StartTraversal() ;

	for ( tmpMach = aRunningJob->runningMachList.Element() ;
	      tmpMach ; tmpMach = !aRunningJob->runningMachList.Advance() ? 
	      			   aRunningJob->runningMachList.Element() : NULL )
	{
		if ( tmpMach->entryList.IsEmpty() )
			continue ;

		fJobLog << "\t" << tmpMach->machPtr->ipAddress << "\t" 
				<< tmpMach->initialPrsrAllocated << "\t" ; 

		tmpMach->entryList.StartTraversal() ;

		RunningEntry_t*	tmpEntry = NULL ; 

		for ( tmpEntry = tmpMach->entryList.Element() ;
	      	      tmpEntry ; tmpEntry = !tmpMach->entryList.Advance() ? 
	      	                 	     tmpMach->entryList.Element() : NULL )
		{
			if ( tmpEntry->processList.IsEmpty() )
				continue ;

			ProcessInfo_t*	tmpProcess = NULL ;

			for ( tmpProcess = tmpEntry->processList.Element() ;
	      	              tmpProcess ; tmpProcess = !tmpEntry->processList.Advance() ? 
	      	                 	                 tmpEntry->processList.Element() : NULL )
			{

				fJobLog << tmpProcess->compTime << "\t" ;
			}
		}

		fJobLog << endl ; 
	}

	fJobLog.flush() ;

	return 0 ; 	
}

// ----------------------------------------------------------------------------------
// ProcessTerminate -- This function updates the appropriate data structures
//			when the GSP receives a MSG_TERMINATE message.
//		      It takes three  parameters : jobId, ip Address of the
//			machine the message is received from, time taken for
//			a process of the job to complete in the machine.
// 		      This function updates the runningJobList to update the
//			appropriate entry for the process terminated,
//			releases the particular processor in the appropriate
//			machine.
// ----------------------------------------------------------------------------------

int
ProcessTerminate( int jobId, const char* ipAddress, int entryNum, struct timeb& timeTakenToComplete )
{
	assert( ipAddress != NULL ) ;

	assert( jobId > 0 ) ;

	// Search the jobList in order to find the entry for the particular job
	Job_t* parCompJob  = Search ( jobList, jobId ) ;

	assert( parCompJob != NULL ) ;

	// Get the running job entry for this job to update the time and free the completed
	// process node
	RunningJob_t*	parCompRunningJob = parCompJob->runEntryPtr ; 

	assert( parCompRunningJob != NULL ) ;

	if ( parCompRunningJob->numCompleted == 0 )
		ftime( &(parCompJob->jobEarliestCompletionTime) ) ;

	// Increment the number of completed processes for this job 
	parCompRunningJob->numCompleted++ ; 
	
	// Find the lsp Machine Entry 
	MachInfo_t* 	lspMachine = parCompRunningJob->Search( ipAddress ) ;

	assert( lspMachine != NULL ) ;

	assert( lspMachine->machPtr != NULL ) ;

	// Reduce the number of processor allocated for this job by 1
	lspMachine->prsrAllocated-- ;

	// Increase the number of free processors in the system
	machineList.processorsFree++ ;

	// Increase the number of free processors on the node by 1 
	lspMachine->machPtr->numProcessorsFree++ ;

	// Free the job pointer in the machine list if the terminated 
	// process is the last process on that machine on behalf of this job
	if ( lspMachine->prsrAllocated == 0 )
	{
		lspMachine->machPtr->DeleteJob( jobId ) ;
	}

	// Access the running entry for the terminated process 
	RunningEntry_t*	runningEntry = lspMachine->Search( entryNum ) ; 

	assert( runningEntry != NULL ) ;

	// Insert the completion details of the process  
	runningEntry->Insert( entryNum, timeTakenToComplete ) ; 
	
	// Check whether the job is fully completed or not 
	if ( parCompRunningJob->numCompleted == parCompJob->numNodesRequired ) 
	{
		// If completed spit out the data and delete the job configuration
		// file	

		ftime( &(parCompJob->jobLatestCompletionTime) ) ; 

		ProcessCompleted( parCompJob, parCompRunningJob ) ;   

		unlink( parCompJob->jobConfigFile.chars() ) ; 

		// Delete the running instance of this job from the list 
		runningList.FindAndDelete( parCompRunningJob ) ; 

		// Free the memory for running instance 
		// by implicitly calling destructors
		delete parCompRunningJob ;

		// Delete the job entry from the jobList  
		jobList.FindAndDelete( parCompJob ) ; 

		// Free the memory for the job  
		delete parCompJob ; 	// Implicitly frees the memory for the entries

		return 0 ; 
	}

	return 0 ; 
}
