#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <ctype.h>
#include <signal.h>
#include <String.h>

// Socket related include files
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/time.h>
#include "List.h"
#include "LSP.h"

// Externals
extern oo_List<LSPStat_t>	jobDetails ;

// Locals 

JobTable	LocalHostJobs ;		// Table that maps job ids to process ids
char		GSPipAddress[MAXHOSTNAMELEN] ; // IP Address of GSP in dot notation 
char		MyIPAddress[MAXHOSTNAMELEN] ; // IP Address of LSP in dot notation 

String		HostConfigFile ; 	// Name of the host configuration file
int		GSPportNumber ;		// Port Number on which the GSP is listening 
int		LSPportNumber ;		// Socket and port number on which the 
int		LSPServerSock ;		// LSP is listening to provide service. 
int		prsrNode = 0 ;		// Number of processors at this node 

int
SendStartupMessage(int numProcessors)
{
	// Create a message 	
	MessHeader_t message(MyIPAddress) ; 

	// Create a socket and connect to GSP 
	int clientSock = CreateClientSocket( GSPipAddress, GSPportNumber ) ; 
	
	if ( clientSock < 0 )
		return -1 ;		

	message.ConstructStartupMessage( numProcessors ) ;

	if ( message.Send( clientSock ) < 0 )
	{
		cerr << "Sending Startup Message Failed " << sys_errlist[errno]
		     << endl ; 

		close(clientSock) ;

		return -1 ;
	}
		
	close(clientSock) ;

	return 0 ;
}

int
SendWithdrawMessage(int numProcessors)
{
	// Check whether any process in the system currently
	// running in this node or not 
	if ( !jobDetails.IsEmpty() )
		return -1 ;
		
	// Create a message 	
	MessHeader_t message(MyIPAddress) ; 

	// Create a socket and connect to GSP 
	int clientSock = CreateClientSocket( GSPipAddress, GSPportNumber ) ; 
	
	if ( clientSock < 0 )
		return -1 ;		

	message.ConstructWithdrawMessage( numProcessors ) ;

	if ( message.Send( clientSock ) < 0 )
	{
		cerr << "Sending Withdraw Message Failed " << sys_errlist[errno]
		     << endl ; 

		close(clientSock) ;

		return -1 ;
	}
		
	close(clientSock) ;

	return 0 ;
}

int
Initialize(const char* GSPhostName)
{
	char* myName = NULL ;

	// Initialize the GSP IP address 
	if ( GetIPAddress( HostConfigFile.chars(),GSPhostName, GSPipAddress ) < 0 )
	{
		// AMIT what about an error message here ?
		exit(1) ;
	}

	// Now get my host name  
	if ( !(myName = GetHostName()) )
	{
		cerr << "Problem getting my host name. " << endl ;
		exit(1) ;
	}

	// Now get my ip address. this will succeed only if there is
	// an entry for this machine in configuration file 
	if ( GetIPAddress( HostConfigFile.chars(), myName, MyIPAddress ) < 0 )  
		exit(1) ;

	// Get the number of processors for me from the configuration file
	int numProcessors = GetConfigurationInfo( HostConfigFile.chars(), MyIPAddress ) ;

	if ( numProcessors <= 0 )
	{
		cerr << "Entry for number of processors is invalid for host "
		     << myName << endl ;
		exit(1) ;
	}

	prsrNode = numProcessors ;

	// Create the server socket to listen to
	LSPServerSock = CreateServerSocket( LSPportNumber ) ;

	if ( LSPServerSock < 0 )
	{
		cerr << "Unable to create server socket to listen in port " 
		     << LSPportNumber << endl ;
	
		return -1 ; 
	}

	// Send startup Message advertising my arrival
   	if ( SendStartupMessage( numProcessors ) < 0 )
	 	return -1 ;

	return 0 ;
}

		
// Before issuing the KILL signal to the child process set a flag on the job table 
// to indicate that the job is being killed. So that once all the processes
// belonging to the job gets killed an ack is send back to the GSP informing
// that the job has been killed successfully. 
// Rather than invoking SIGCHLD 

void
SignalHandler_CHLD(int x)
{
	// Save the errno 
	int b_errno = errno ;

	// The for loop is used to fix a problem that occurs when
	// child processes terminate too fast, and cause messages
	// to be lost.
   
	int status ;	// To get the status of the terminated child process

	int childPid ;		
	
	signal_lock() ;

	for ( ; ; )	
	{	
		childPid = wait3( &status, WNOHANG, NULL ) ;
	
		if ( childPid <= 0 )
		{
			//Restore the errno
			errno = b_errno ;

			signal_unlock() ;

			return ;
		}
	
		cerr << "Inside signal handler" << endl ;

		SendTerminatedMessage( childPid ) ;

		cerr << "Leaving signal handler" << endl ;
	}
}
	
void
SignalHandler_INT(int x)
{
	close(LSPServerSock) ;

	exit(1) ;
}

int
main( int argc, char* argv[] )
{
	extern	int	optind ;
	extern	char*	optarg ;

	char*	gspHostName = NULL ;
	char*	gspPortNumber = NULL ;
	char*	myPortNumber = NULL ;
	char*	conFile = NULL ; 
	char*	myName = NULL ;
						
	int	c ;

	while ( (c = getopt(argc,argv,"f:s:p:l:") ) != EOF )
	{
		 switch(c)
		 {
			case 'f' :
				 conFile = optarg ;
				 break ;

			case 's' :
				 gspHostName = optarg ;
				 break ;

			case 'p' :
				 gspPortNumber = optarg ;
				 break ;
			
			case 'l' :
				 myPortNumber = optarg ;
				 break ;

			case '?' :
			         cerr << "Usage : " << argv[0] << 
				 " -s <GSP Host Name> [ -p <GSP Port Number> -l <Local LSP Port Number> "
				 << "-f <Configuration File> ]" << endl ;
				 exit(1) ;
		 }
	}

	if ( !gspHostName )
	{
		cerr << "Usage : " << argv[0] << 
			" -s <GSP Host Name> [ -p <GSP Port Number> -l <Local LSP Port Number> "
		     << "-f <Configuration File> ]" << endl ;
		exit(1) ;
	}	
	else if ( !OurGetHostEntry(gspHostName) )
	{
		cerr << gspHostName << " does not exist. " << endl ;
		exit(1) ; 
	}

	// Validate the GSP port number 
	if ( (GSPportNumber = ValidatePortNumber(gspPortNumber,GSP_PORT_NUMBER)) < 0 )
	{
		cerr << "GSP Port Number should be a valid integer" << endl ;		
		exit(1) ;
	}

	// Validate the LSP port number 
	if ( (LSPportNumber = ValidatePortNumber(myPortNumber,LSP_PORT_NUMBER)) < 0 )
	{
		cerr << "LSP Port Number should be a valid integer" << endl ;		
		exit(1) ;
	}

	// Validate the config file 	
	if ( !conFile )
	{
		HostConfigFile = ConfigFile ;
	}
	else if ( !CheckFileExistence( conFile ) )
	{
		HostConfigFile = conFile ;
	}
	else
	{
		cerr << conFile << " does not exist. Quitting..." << endl ;		
		exit(1) ;
	}
		
	// Now call the initialize function to initialize the parameters  
	// and send the startup message

	if ( Initialize(gspHostName) < 0 )
	{
		cerr << "LSP Initialization Failed. Quitting..." << endl ;
		exit(1) ; 
	}

	struct sockaddr_in	clientAddr ;

	int			clientAddrLen ;

	signal( SIGCHLD, SignalHandler_CHLD ) ;
	signal( SIGINT, SignalHandler_INT ) ;

	fd_set readset ;

	while (1)
	{
		FD_ZERO( &readset ) ;

		FD_SET( LSPServerSock, &readset ) ;

		cout << LSPServerSock << endl ; 

		cout << "Before select " << endl ;
		
		int s = select( LSPServerSock + 1, &readset, NULL, NULL, NULL ) ;

		cout << "After select s = " << s << endl ;

		// signal_lock() ;

		if ( s > 0 && FD_ISSET( LSPServerSock, &readset ) ) 
		{
			signal_lock() ;

			int clientD = accept( LSPServerSock,(struct sockaddr *)&clientAddr, &clientAddrLen) ;

			if (clientD < 0 )
			{
				cerr << "Error in accept:" << sys_errlist[errno] << endl ; 

				continue ;
			}	

			signal_unlock() ;

			MessHeader_t msg(MyIPAddress) ;

			if ( msg.Receive(clientD) < 0 ) 
			{
				cerr << "Error in receiving message header " 
				     << sys_errlist[errno] << endl ;

				close(clientD) ;

				continue ; 
			}	

			cout << msg << endl ;

			if ( msg.msgType == MSG_SCHEDULE )
			{
				// Verify whether it is GSP sending the schedule message 

				char* buffer = new char [msg.msgLen+2] ; 

				bzero( buffer, sizeof(char) * (msg.msgLen + 2) ) ; 
				
				// Error checking  
				read(clientD,buffer,msg.msgLen) ;  

				cout << buffer << endl ;

				close(clientD) ; 

				String msgBuffer(buffer) ; 

				free(buffer) ; 

				SplitMessageAndSchedule( msg.jobId, msgBuffer) ;
			}
			else if ( msg.msgType == MSG_KILL )
			{
				if ( KillJob(msg.jobId) < 0 )
				{
					cerr << "Killing Job Id" << endl ;
				}
			}
			else if ( msg.msgType == MSG_QUERY )
			{
				if ( msg.jobId >= 0 )
				{	
					SingleJobQuery( clientD, msg.jobId, MyIPAddress ) ;  
				}
				else if ( msg.jobId < 0 )  
				{
					AllJobQuery( clientD, MyIPAddress ) ;  	
				}
			}
			else if ( msg.msgType == MSG_SUSPEND )
			{
				if ( msg.jobId >= 0 )
				{
					ResumeORSuspendJob( msg.jobId, SUSPEND ) ;
				}
			}
			else if ( msg.msgType == MSG_RESUME )
			{
				if ( msg.jobId >= 0 )
				{
					ResumeORSuspendJob( msg.jobId, RESUME ) ;
				}
			}			
			else if ( msg.msgType == MSG_STARTUP )
			{
				SendStartupMessage( prsrNode ) ;
			}	
			else if ( msg.msgType == MSG_WITHDRAW )
			{
				SendWithdrawMessage( prsrNode ) ;
			}

			close(clientD) ; 
		}
		else if ( s < 0 )
		{
			cerr << "Error in select:" << sys_errlist[errno] << endl ;

			continue ;
		}

		// signal_unlock() ;
	}

			

	/*
	for ( int i = 1 ; i <= 2 ; i++ )
	{
		// Block the SIGCHLD until fork and updates of the tables complete 
		int oldmask = sigblock( sigmask(SIGCHLD) ) ;

		int pid = fork() ;

		cout << pid << endl ;

		// Parent process
		if ( pid )
		{
			LocalHostJobs.InsertJobId( i ) ;

			LocalHostJobs.InsertProcessId( i, pid ) ;
		}
		else if ( pid == 0 )
		{
			// Restore the mask in the child 
			// sigsetmask(oldmask) ;

			// child process 
			if ( execl("/u/grad/ramasamy/736/code/c/test","test",(char *) 0 ) < 0 )
			{
				cerr << sys_errlist[errno] << endl ;
			
				exit(1) ;
			}

		}

		// Restore the mask in the parent 
		sigsetmask(oldmask) ;
	}

	sleep(30) ;


	for ( i = 1 ; i <=2 ; i++ )
	{
		KillJob(i) ;
	}	

	while(1)
	{
	}

	*/
}
