# ifndef  _OO_LIST_H_
# define  _OO_LIST_H_

/************************************************************
 * Include Headers
 ************************************************************/
# include <iostream.h>

# ifndef NULL
# define NULL       0
# endif

/************************************************************
 *  Type Definitions
 ************************************************************/


/************************************************************
 *   class oo_Link
 *   o  This is a private class used by the oo_List class.
 *     Each element of the list is stored as an instance of 
 *         this oo_Link class.
 *     the class oo_List has exclusive access to the members of
 *        this class.
 ************************************************************/
template <class ListElementType>
class  oo_Link  {

friend class oo_List<ListElementType>;

private:
  ListElementType * m_pValue;       // actual element is stored here
  oo_Link<ListElementType> * m_pNext; // next element in the  list
  
// Note that this is private class, Only oo_List<ListElementType> can use it.
private:

  // Constructor
  oo_Link(ListElementType * pElement = NULL,
	  oo_Link<ListElementType> * pNewNext = NULL)
    {  m_pValue = pElement;
       m_pNext = pNewNext;
     }
};  // class oo_Link


/************************************************************
 *   class oo_List
 *   o  A List is implemented as a chain of circularly linked
 *       instances of class oo_Link.
 *     A HEAD pointer always points to the first element of list.
 *     A CURRENT pointer is used for traversal through the list
 *    List is EMPTY if HEAD == NULL
 ************************************************************/
template <class ListElementType>
class  oo_List  {
private:
  oo_Link<ListElementType> * m_pHead;      // pointer to head node.
  oo_Link<ListElementType> * m_pCurrent;   // used for traversal of list.

public:
  //   Defintions of the functions

  /******** Constructor functions *******/
  oo_List()       { m_pHead = NULL;}       // Default Constructor
  
  oo_List(ListElementType * pElement)
    {
      m_pHead = m_pCurrent = new oo_Link<ListElementType> (pElement);
      m_pHead->m_pNext = m_pHead;    // establish the circular link
    }

  /******* Checking list info *********/
  /**************************************************
   *  IsEmpty(): Checks if the list is empty
   *  RETURNS:
   *     NON-ZERO : if the list is empty
   *     ZERO if the list is not empty
   **************************************************/
  int IsEmpty(void)              
    { return m_pHead == NULL; }


  /******* Functions for Traversal *******/
  /**************************************************
    *  StartTraversal(): Begins a traversal of the list
    *  RETURNS:
    *    returns a NON-ZERO value if list is empty
    *    ZERO : if list is not empty
    * Note: 
    *   At any time only one traversal is permitted
    *   Always call this function before a traversal
    **************************************************/
  int StartTraversal(void) 
    { return  (m_pCurrent = m_pHead) == NULL; }

  /**************************************************
    *  Element(): Returns the current element of list
    *  Returns:
    *      NULL  if list is empty
    *      non-NULL if an element is present
   **************************************************/
  ListElementType * Element(void)
    {
      if (IsEmpty())
	return NULL;
      else
	return m_pCurrent->m_pValue;
    }

  /**************************************************
    *  Advance() :  Used in a traversal to
    *    advance the CURRENT pointer to next element.
    * RETURNS:
    *   NON_ZERO if an attempt is made to advance past
    *       the last element of a list
    *   ZERO on a successful advance.
    * Note:
    *   It is an error to advance on an empty oo_List
   **************************************************/
  int Advance(void)
    {
      if (IsEmpty())
	{ cerr << "ERROR: oo_List::Advance() : oo_List is empty\n";
	  return 1;
	}
      return (m_pCurrent = m_pCurrent->m_pNext) == m_pHead; 
                        // Non-zero if end of list
    }


   /**************************************************
     *  Insert(): 
     *   Inserts a new element before the current element
     *      and make that as current element.
     *   If List is empty , then make the new element as
     *      as well as the current element
     * RETURNS:
     *   the modified oo_List object
    **************************************************/
  oo_List<ListElementType> & Insert(ListElementType * pNewElement)
    {
      if (IsEmpty())
	{
	  m_pHead = m_pCurrent = new oo_Link<ListElementType> (pNewElement);
	  m_pHead->m_pNext = m_pHead;    // establish the circular link
	}
      else
	{
	  /* Insert a new link after the current element.
	     Copy the contents of the current element into the new link.
	     and then replace the contents of current element by the new
	     element
	     */
	  m_pCurrent->m_pNext = 
	    new oo_Link<ListElementType> (m_pCurrent->m_pValue, 
					  m_pCurrent->m_pNext);
	  m_pCurrent->m_pValue = pNewElement;
	}
      return *this;
    }

  /**************************************************
   *  InsertAtEnd()
    *    Insert a new element at the end of the list
    *  If the list is empty , then make new element
    *    first as well as the current element of list.
    * RETURNS:
    *   the modified oo_List object
    **************************************************/
  oo_List<ListElementType> & InsertAtEnd(ListElementType * pNewElement)
    {
      if (IsEmpty())
	{
	  m_pHead = m_pCurrent = new oo_Link<ListElementType> (pNewElement);
	  m_pHead->m_pNext = m_pHead;    // establish the circular link
	}
      else
	{
	  /* Insert a new link after the head element.
	     Copy the contents of the head to the new link.
	     Replace the contents of head element by new element
	     Finally, move the head to the new link
	     */
	  m_pHead->m_pNext =
	    new oo_Link<ListElementType> (m_pHead ->m_pValue,
				       m_pHead ->m_pNext);
	  m_pHead->m_pValue = pNewElement;
	  if (m_pCurrent == m_pHead)
	    m_pCurrent = m_pCurrent->m_pNext;
	  m_pHead = m_pHead ->m_pNext;
	}

      return *this;
    }

  /**************************************************
    *  DeleteAndAdvance()
    *    Delete the current element and advance the
    *    CURRENT pointer to the next element.
    * RETURNS:
    *   ZERO if the last element is deleted
    *  NON-ZERO otherwise
    * Note: An Error to use it on empty list
   **************************************************/
  ListElementType* DeleteAndAdvance(void)
    {
      int fDeletingFirstElem = 0;

      if (IsEmpty())
	{
	  cerr << "ERROR: oo_List::DeleteAndAdvance() : oo_List is empty\n";
	  return 1;
	}
      
      if (m_pHead == m_pCurrent)
	fDeletingFirstElem = 1;   // we are deleting first element

      oo_Link<ListElementType> * pTmp = m_pCurrent->m_pNext;
      
      if (pTmp == m_pHead)
	if (pTmp->m_pNext == m_pHead)
	  {
	    // Deleting the only element in the list
	    delete pTmp;
	    m_pHead = m_pCurrent = NULL;
	    return 1;       // List is empty now
	  }
      else // We are deleting the last element in list
	m_pHead = m_pCurrent;
      // When this function completes both Head and Current will
      // point to the same node (which will be new head of list)
      
      // Copy over values from node next to the CURRENT
      m_pCurrent->m_pValue = pTmp ->m_pValue;
      m_pCurrent->m_pNext = pTmp->m_pNext;

      // Delete the next node
      delete pTmp;
      
      if (m_pHead == m_pCurrent && !fDeletingFirstElem)
	return 1;      // we deleted the last element
      else
	return 0;      // we did not delete the first element
    } // DeleteAndAdvance()

  /**************************************************
   *  Replace() 
   *    Replace the current element by the new element
   *  Note: This is an error if the list is empty
   **************************************************/
  oo_List<ListElementType> & Replace(ListElementType * pNewElement)
    {
      if (IsEmpty())
	cerr << "ERROR: oo_List::Replace() : oo_List is empty\n";
      else
	m_pCurrent->m_pValue = pNewElement;
      // Freeing the old element is left to overloaded assignment operator
      // or the garbage collector

      return *this;
    }

  /**************************************************
    * Find()
    *  Find the first occurence of the given element
    *    
    * RETURNS:
    *   ZERO: if an element is not found
    *   NON-ZERO: on successful
   **************************************************/
  ListElementType* Find(ListElementType * pElement)
    {
      for ( int fEndofList = StartTraversal();
	   !fEndofList; fEndofList = Advance())
	if (Element() == pElement)     // only pointer comparison is done
	  {
	    return Element();
	  }
      
      return NULL;
    }

  /**************************************************
    * FindAndDelete()
    *  Find the first occurence of the given element
    *    and delete it from the list
    * RETURNS:
    *   ZERO: if an element is not found
    *   NON-ZERO: on successfull deletion
   **************************************************/
  int FindAndDelete(ListElementType * pElement)
    {
      for ( int fEndofList = StartTraversal();
	   !fEndofList; fEndofList = Advance())
	if (Element() == pElement)     // only pointer comparison is done
	  {
	    DeleteAndAdvance();
	    return 1;
	  }
      
      return 0;
    }

  /**************************************************
    * FindNext()
    *  Find the first occurence of the given element
    *    and return the next element  
    * RETURNS:
    *   NULL: if the given element is the last element found
    *   Element: Otherwise 
   **************************************************/
  ListElementType* FindNext(ListElementType * pElement) 
    {
	
      for ( int fEndofList = StartTraversal() ;
	    !fEndofList ; fEndofList = Advance() )
	if (Element() == pElement)     // only pointer comparison is done
	  {
	   	break ;  
	  }
      if ( Advance() ) 
	return NULL ; 
      
      return Element() ;
    }	

      
	   
   /**************************************************
     *  DeleteList()
     *   Deletes the whole list
     * RETURNS:
     *    None
    **************************************************/
  void DeleteList(void)
    {
      for (int fEndofList = StartTraversal();
	   !fEndofList;
	   fEndofList = DeleteAndAdvance())
	;
    }

  /**************************************************
    *  Length()
    *    Finds the length of given list
    * RETURNS:
    *     an integer value
    **************************************************/
  int Length(void)
    {
      int length = 0;

      for(int fEndofList = StartTraversal();
	  !fEndofList;
	  fEndofList = Advance())
	length ++;
      
      return length;
    }
  
  /**************************************************
    *  ListToArray()
    *    Converts the given list into an array of N elements
    *     and initialises the array with the elements of the list.
    *  RETURNS:
    *    The array containing the pointers to elements
    *     present in the array
    **************************************************/
  ListElementType ** ListToArray(int * pArraySize = 0)
    {
      int length = Length();
      
      if (pArraySize)
	*pArraySize = length;

      ListElementType ** ppElement = new ListElementType * [length];

      int index;
      int fEndofList;
      for ( index = 0, fEndofList = StartTraversal();
	   !fEndofList;
	   fEndofList = Advance())
	ppElement[index++] = Element();
      
      return ppElement;
    }

  /**************************************************
    *  Append()
    *   Appends one oo_List to another
    *  RETURNS:
    *   NONE (the first list is appended with the second one
    **************************************************/
  void Append(oo_List * pList2)
    {

      if (IsEmpty())      // first list is empty
	m_pHead = m_pCurrent = pList2->m_pHead;
      else
	if (!pList2->IsEmpty())
	  {
	    // Both lists are non-empty

	    // Find the last element of the first list
	    for (oo_Link<ListElementType> * pTmp = m_pHead;
		 pTmp->m_pNext != m_pHead;
		 pTmp = pTmp->m_pNext)
	      ;
	    
	    // Link the two lists
	    pTmp->m_pNext = pList2->m_pHead;
	    
	    // Find the last element of the second list
	    for (pTmp = pList2->m_pHead;
		 pTmp->m_pNext != pList2->m_pHead;
		 pTmp = pTmp->m_pNext)
	      ;
	    
	    // Link the list circularly
	    pTmp->m_pNext = m_pHead;
	  }
    } // Append()

  /**************************************************
   *  FirstElement()
   *     Get the first element of the list
   *  RETURNS:
   *    a pointer to first element of the list
   *   NULL if an empty list
   **************************************************/
  ListElementType * FirstElement(void)
    {
      if (m_pHead) 
	return m_pHead->m_pValue;
      else
	return NULL;
    }

  /**************************************************
    *  NthElement()
    *    Get the Nth Element of the list
    *  RETURNS:
    *    a pointer to Nth element of the list
    *   NULL if an empty list
    *  Note: 
    *     Validity of N is not checked for.
    **************************************************/
  ListElementType * NthElement(int N)
    {
      oo_Link<ListElementType> * pTmp = m_pHead;
      
      if (!pTmp)
	return NULL;
      
      while (N-- > 0)
	pTmp = pTmp->m_pNext;
      
      return  pTmp->m_pValue;
    }

//  typedef void (ListElementType:: * PMFCN)(void);

  /**************************************************
   *  ApplyFunction()
   *     apply a given function to all the members of
   *      given list
   * Input:
   *    [pm_fcn] : pointer to a member function 
   *          of the class ListElementType
   * RETURNS:
   *    ZERO  if list is empty
   *    NON-ZERO otherwise
   **************************************************/
//   int ApplyFunction( PMFCN *pm_fcn)
//     {
//       if (IsEmpty())
// 	return 0;
//       else
// 	{
// 	  for(int fEndofList = StartTraversal();
// 	      !fEndofList;
// 	      fEndofList = Advance())
// 	    Element()->(*pm_fcn)();     // make a function call
// 	  return 1;
// 	  }
//    } // ApplyFunction

};  // class oo_List



# endif   /* _OO_LIST_H_ */

/*********************** End of File ***********************/
